/*
 * Copyright (c) 2003, 2004 Adaptec Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer,
 *    without modification.
 * 2. Redistributions in binary form must reproduce at minimum a disclaimer
 *    substantially similar to the "NO WARRANTY" disclaimer below
 *    ("Disclaimer") and any redistribution must be conditioned upon
 *    including a substantially similar Disclaimer requirement for further
 *    binary redistribution.
 * 3. Neither the names of the above-listed copyright holders nor the names
 *    of any contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * Alternatively, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") version 2 as published by the Free
 * Software Foundation.
 *
 * NO WARRANTY
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 */
#include <linux/module.h>
#include <linux/config.h>
#include <linux/genhd.h>
#include <linux/list.h>
#include <linux/raid/md.h>
#include "md_metadata_ddf.h"

#define	DDF_BUFSIZ	PAGE_SIZE
#define DDF_NORMAL_BID	10
#define MD_DDF_MAX_PDE 15
#define MD_DDF_MAX_VDE 7

struct md_ddf_array_info {
	struct list_head		ddf_array_links;
	struct ddf_vd_record		*vdr;
	struct ddf_vd_entry		*vde;
	struct ddf_vdc_record		*vdcr;
	uint32_t			*pds_array;
	uint32_t			*ss_array;
	mddev_t				*mddev;
	uint32_t			DDF_Sequence_Number;
	int				needed_disks;
};

struct md_ddf_disk_info {
	struct list_head		ddf_disk_links;
	struct block_device		*blkdev;
	mdk_rdev_t			*rdev;
	struct md_ddf_array_info	*parent;

	/* Pointers to things in the metadata that are hard to find */
	struct ddf_pd_entry		*pde;
	int				ss_len;
	int				pds_len;

	/* Here we keep a copy of the DDF info */
	struct ddf_header		*hdr_anchor;
	struct ddf_header		*hdr_pri;
	struct ddf_cd_record		*cd;
	struct ddf_pd_record		*pdr;
	struct ddf_vd_record		*vdr;
	uint8_t				*cr;	/* This holds the storage
						 * for the following three
						 * pointers.
						 */
	struct ddf_vdc_record		*vdcr;
	struct ddf_vuc_record		*vuc;
	struct ddf_sa_record		*sa;
	struct ddf_pdd_record		*pdd;
	struct ddf_bbm_log		*bbm;
	struct ddf_vendor_log		*vendor;

	uint32_t			ddf_dirty_mask;
#define DI_DIRTY_HDR_ANCHOR	0x0000
#define DI_DIRTY_HDR_PRI	0x0001
#define DI_DIRTY_HDR_SEC	0x0002
#define DI_DIRTY_CD		0x0004
#define DI_DIRTY_PDR		0x0008
#define DI_DIRTY_VDR		0x0010
#define DI_DIRTY_VDCR		0x0020
#define DI_DIRTY_VUC		0x0040
#define DI_DIRTY_SA		0x0080
#define DI_DIRTY_PDD		0x0100
#define DI_DIRTY_BBM		0x0200
#define DI_DIRTY_VENDOR		0x0400
};

static void ddf_dumpblock(void *, int) __attribute__((unused));
static int md_dev_op_ddf(dev_t dev, mdk_dev_op_t, void *arg);
static int md_member_op_ddf(mdk_member_t *member, int, void *arg);
static int ddf_claim(dev_t dev, void *arg);
static int ddf_get_anchor(struct block_device *blkdev, struct ddf_header **hdr);
static int ddf_attach(dev_t dev, void *arg);
static int ddf_add_spare(struct md_ddf_disk_info *di);
static int ddf_get_array_info(struct md_ddf_disk_info *di,
			      struct md_ddf_array_info **ai);
static int ddf_get_disk_info(struct block_device *blkdev,
			     struct ddf_header *anchor_header,
			     struct md_ddf_disk_info **di);
static int ddf_attach_drive(struct md_ddf_disk_info *di);
static mddev_t *ddf_attach_array(struct md_ddf_array_info *ai);
static int ddf_check_runnable(mddev_t *mddev);
static int ddf_to_md_disk_state(uint8_t pd_state);
static int md_set_disk_state_ddf(mddev_t *mddev, mdk_rdev_t *disk,
				 mdk_mstate_t state);
static int md_set_array_state_ddf(mddev_t *mddev, mdk_mstate_t state);
static uint8_t md_to_ddf_disk_state(mdk_mstate_t state);
static uint8_t md_to_ddf_array_state(mdk_mstate_t state);
static void ddf_free_disk_info(struct md_ddf_disk_info *);

static mdk_metapersonality_t md_ddf_pers = {
	.name =		DDF_METADATA_NAME,
	.owner =	THIS_MODULE,
	.dev_op =	md_dev_op_ddf,
	.member_op =	md_member_op_ddf,
	.refcount =	ATOMIC_INIT(1),
};

static LIST_HEAD(diskqueue);
static LIST_HEAD(arrayqueue);

/* Convienience routine to dump a block in hex */
static void
ddf_dumpblock(void *data, int len)
{
        int i, j;

        for (i = 0; i < len; i++) {
                printk("\noffset= %03d: ", i);
                for (j = 0; j < 8; j++) {
                        if (i + j == len)
                                break;
                        printk("0x%02x ", (uint8_t)((uint8_t*)(data))[i + j]);
                }
                i += j - 1;
        }
        printk("\n");
}

/* Entry points into the module */
static int
md_dev_op_ddf(dev_t dev, mdk_dev_op_t op, void *arg)
{
	switch (op) {
	case MDK_DEVOP_CLAIM:
		return (ddf_claim(dev, arg));
	case MDK_DEVOP_ATTACH:
		return (ddf_attach(dev, arg));
	case MDK_DEVOP_INITIALIZE:
	case MDK_DEVOP_WIPE_METADATA:
	default:
		return (-EINVAL);
	}

	return (0);
}

static int
md_member_op_ddf(mdk_member_t *member, int op, void *arg)
{
	switch (op) {
	case MDK_MEMBER_OP_GET_CHECKPOINT:
	{
		mdk_checkpoint_arg_t *cp;

		cp = (mdk_checkpoint_arg_t *)arg;
		cp->checkpoint = 0;
		return (0);
	}
	case MDK_MEMBER_OP_SET_CHECKPOINT:
		return (0);
	case MDK_MEMBER_OP_SET_STATE:
	case MDK_MEMBER_OP_SET_ARRAY_ACTION:
	case MDK_MEMBER_OP_SET_USER_ACTION:
	case MDK_MEMBER_OP_ALLOC_SPARE:
	default:
		return (-EINVAL);
	}

	return (0);
}

/* All that we want to do here is see if a valid anchor exists.  No other
 * actions are needed, nor can we save away any state.
 */
static int
ddf_claim(dev_t dev, void *arg)
{
	struct block_device *blkdev;
	struct ddf_header *header;
	int error;
	int *bid;

	bid = (int *)arg;

	blkdev = md_get_blkdev(dev);
	if (blkdev == NULL) {
		printk("DDF: md_get_blkdev returned NULL\n");
		return (-EINVAL);
	}

	error = ddf_get_anchor(blkdev, &header);
	md_release_blkdev(blkdev);
	if (error != 0)
		*bid = 0;
	else
		*bid = DDF_NORMAL_BID;

	if (header != NULL)
		kfree(header);

	return (0);
}

/* Return the DDF anchor record from a given drive. This function allocates
 * the buffer, but the caller needs to free it.  Since the claim code needs
 * to know the difference between a failure to get the block and the block
 * not having the right DDF info, only worry about returning status for the
 * first case.  The callers will each have to examine the contents.
 */
static int
ddf_get_anchor(struct block_device *blkdev, struct ddf_header **hdr)
{
	struct ddf_header *header;
	uint64_t ddf_anchor_base;

	/* Find the end of the disk, which holds Anchor DDF block */
	ddf_anchor_base = get_capacity(blkdev->bd_disk) - 1;

	*hdr = NULL;
	header = (struct ddf_header *)kmalloc(DDF_BUFSIZ, GFP_KERNEL);
	if (header == NULL) {
		printk("md_metadata_ddf: failed to allocate buffer\n");
		return (-ENOMEM);
	}
	memset(header, 0, DDF_BUFSIZ);

	/* Get the anchor block and validate it */
	/* XXX Should the blocksize be cached and restored?  Is this the
	 * right place to set it?
	 */
	if (md_read_blocks(blkdev, ddf_anchor_base, 1, (uint8_t *)header)!= 0) {
		printk("ddf_get_anchor: cannot read anchor block\n");
		kfree(header);
		return (-EIO);
	}

	if ((le32_to_cpu(header->Signature) != DDF_HEADER_SIGNATURE) ||
	    (le32_to_cpu(header->Header_Type) != DDF_HEADER_ANCHOR)) {
		kfree(header);
		return (-EINVAL);
	}

	*hdr = header;
	return (0);
}

/*
 * Do the work of fetching all of the records and doing fun things
 * with them.
 */
static int
ddf_attach(dev_t dev, void *arg)
{
	struct block_device *blkdev;
	struct ddf_header *header;
	struct md_ddf_disk_info *di;
	int error;

	di = NULL;
	header = NULL;

	blkdev = md_get_blkdev(dev);
	if (blkdev == NULL) {
		printk("DDF: md_get_blkdev returned NULL\n");
		return (-EINVAL);
	}

	/* Start by getting the archor header */
	if ((error = ddf_get_anchor(blkdev, &header)) != 0)
		return (error);

	/* Next get the DDF records for the disk_info structure */
	if ((error = ddf_get_disk_info(blkdev, header, &di)) != 0) {
		kfree(header);
		return (error);
	}

	/* Is this drive a spare? */
	if ((le16_to_cpu(di->pde->PD_Type) &
	    (DDF_PDE_GLOBAL_SPARE | DDF_PDE_CONFIG_SPARE)) != 0)
		return (ddf_add_spare(di));
	else
		return (ddf_attach_drive(di));
}

static int
ddf_attach_drive(struct md_ddf_disk_info *di)
{
	struct ddf_pd_entry *pde;
	struct md_ddf_array_info *ai;
	mdk_member_t *rdmember, *mdmember;
	mdk_rdev_t *rdev;
	mddev_t *mddev;
	uint32_t pdseq;
	int state, error, i, pds_index;

	/* Create an rdev member object for the disk */
	if ((rdev = md_rdev_alloc(NULL, MDK_MSUBTYPE_CHILD, di->blkdev,
				  MDK_PRIORITY_UNKNOWN)) == NULL) {
		printk("ddf_attach_drive: cannot alloc md_disk\n");
		return (-EINVAL);
	}

	rdmember = to_member(rdev);
	rdmember->data_size = (sector_t)le64_to_cpu(di->pde->Configured_Size);
	rdmember->data_offset = 0;	/* XXX */
	rdmember->metadata_private = di;
	rdmember->state = ddf_to_md_disk_state(di->pde->PD_State);

	/* Is the drive a valid part of an array? */
	if (((le16_to_cpu(di->pde->PD_Type) & DDF_PDE_PARTICIPATING) == 0) ||
	    ((le16_to_cpu(di->pde->PD_State) &
	    (DDF_PDE_ONLINE | DDF_PDE_REBUILD | DDF_PDE_TRANSITION)) == 0)) {
		return (0);
	}

	/* Now get the array_info structure */
	if ((error = ddf_get_array_info(di, &ai)) != 0)
		return (error);

	/*
	 * Check the sequence numbers to make sure that this member
	 * is in sync with the others.  We should really be more
	 * intelligent here....
	 */
	if ((le32_to_cpu(di->hdr_pri->Sequence_Number) != 0xffffffff) &&
	    (ai->DDF_Sequence_Number != 0xffffffff) &&
	    (le32_to_cpu(di->hdr_pri->Sequence_Number) !=
	    ai->DDF_Sequence_Number)) {
		printk("DDF: Disk %d/%d is out of date, rejecting\n",
		       MAJOR(di->blkdev->bd_dev), MINOR(di->blkdev->bd_dev));
		return (-EINVAL);
	}

	if (le32_to_cpu(di->vdcr->Sequence_Number) != 
	    le32_to_cpu(ai->vdcr->Sequence_Number)) {
		printk("DDF: Disk %d/%d is out of date, rejecting\n",
		       MAJOR(di->blkdev->bd_dev), MINOR(di->blkdev->bd_dev));
		return (-EINVAL);
	}

	/* If the mddev is NULL, the this array is new and hadn't been
	 * initialized yet.
	 */
	if (ai->mddev == NULL)
		mddev = ddf_attach_array(ai);
	else
		mddev = ai->mddev;

	if (mddev == NULL) {
		printk("ddf_attach: ddf_attach_array returned NULL\n");
		return (-EINVAL);
	}

	mdmember = to_member(mddev);

	/* Verify that this drive is actually in the array */
	pde = di->pde;
	pds_index = -1;
	for (i = 0; i < le16_to_cpu(di->vdcr->Primary_Element_Count); i++) {
		pdseq = ai->pds_array[i];
		if (pdseq == le32_to_cpu(pde->PD_Reference)) {
			pds_index = i;
			break;
		}
	}
	if (pds_index == -1) {
		printk("Cannot find entry for drive in VD\n");
		return (-EINVAL);
	}

	/* XXX Adaptec BIOS doesn't always get this field right.
	 * Do this here since we need to look at both array and disk
	 * attributes.
	 */
	if ((rdmember->data_size == 0) && (mddev->level == RAID1))
		rdmember->data_size = mdmember->data_size;

	/*
	 * Add the disk into the array.
	 */
	rdmember->priority = pds_index;
	mddev->raid_disks++;
	mdmember->pers->raid_op(mdmember, MDK_RAID_OP_INIT_MEMBER,
				(void *)rdmember);
	md_member_add_child(mdmember, rdmember);
	mdmember->pers->raid_op(mdmember, MDK_RAID_OP_ADD_MEMBER,
				(void *)rdmember);

	/* See if the array state has changed.  It is initialized to a
	 * non-running state to make sure that it does get started early.
	 * If a state transition happens, then notify the raid module.
	 */
	state = ddf_check_runnable(mddev);
	if (state != mdmember->state) {
		mdmember->state = state;
		mdmember->pers->raid_op(mdmember, MDK_RAID_OP_MSTATE_CHANGED,
					mdmember);
	}

	return (0);
}

static mddev_t *
ddf_attach_array(struct md_ddf_array_info *ai)
{
	mddev_t *mddev;
	mdk_member_t *member;

	/* Can we deal with this array? */
	if ((ai->vdcr->Primary_RAID_Level != 0) &&
	    (ai->vdcr->Primary_RAID_Level != 1)) {
		printk("DDF raid level %d not supported\n",
		       ai->vdcr->Primary_RAID_Level);
		return (NULL);
	}
	if (ai->vdcr->Secondary_Element_Count != 0x01) {
		/* XXX Handle the Adaptec interpretation */
		if (ai->vdcr->Secondary_Element_Count != 0x00) {
			printk("Multilevel DDF arrays not supported\n");
			return (NULL);
		}
	}

	mddev = md_mddev_alloc(&toplevel_arrays, MDK_MSUBTYPE_TOPLEVEL,
			       0 /* order */);
	if (mddev == NULL) {
		printk("Cannot allocate md_array\n");
		return (NULL);
	}

	member = to_member(mddev);
	switch (ai->vdcr->Primary_RAID_Level) {
	case 0:
		mddev->level = 0;
		ai->needed_disks = le16_to_cpu(ai->vdcr->Primary_Element_Count);
		break;
	case 1:
		mddev->level = 1;
		ai->needed_disks = 1;
		break;
	default:
		/* This should not be reached */
		break;
	}

	ai->mddev = mddev;
	mddev->preferred_unit = -1;
	mddev->chunk_size = (1 << ai->vdcr->Stripe_Size) * 512;
	mddev->max_disks =le16_to_cpu(ai->vdcr->Primary_Element_Count);
	member->data_size = (sector_t)le64_to_cpu(ai->vdcr->Size);
	member->data_offset = 0;	/* XXX need a real value here */
	member->metadata_private = ai;
	member->state = MDK_MSTATE_FAILED; /* Initial state */
	member->metapers = &md_ddf_pers;
	member->pers = md_find_pers(mddev->level);
	if (member->pers == NULL) {
		/* No personality was found for this.  Should we tear down
		 * all of the objects and wait for later?
		 */
		member_put(member);
		return (NULL);
	}

	/* Call the raid module to do basic init on the mddev */
	member->pers->raid_op(&mddev->member, MDK_RAID_OP_INIT, NULL);

	return (mddev);
}

/* Probe a disk for the relevant disk-specific DDF records, and return
 * them in a disk_info struct.  The caller is responsible for freeing this.
 */
static int
ddf_get_disk_info(struct block_device *blkdev, struct ddf_header *anchor_header,
		  struct md_ddf_disk_info **disk_info)
{
	struct md_ddf_disk_info *di, *tmp_info;
	uint64_t ddf_primary_base;
	uint32_t pdref;
	uint32_t offset, length;
	uint16_t crl;
	uint8_t *buffer;
	int i;

	*disk_info = NULL;

	di = kmalloc(sizeof(struct md_ddf_disk_info), GFP_KERNEL);
	if (di == NULL) {
		printk("ddf_get_disk_info: cannot allocate disk_info\n");
		return (-ENOMEM);
	}
	memset(di, 0, sizeof(struct md_ddf_disk_info));
	di->blkdev = blkdev;
	di->hdr_anchor = anchor_header;

	/* Find the Primary DDF header block */
	di->hdr_pri = kmalloc(DDF_HEADER_LENGTH, GFP_KERNEL);
	if (di->hdr_pri == NULL) {
		printk("md_metadata_ddf: failed to allocate buffer\n");
		ddf_free_disk_info(di);
		return (-ENOMEM);
	}

	ddf_primary_base = le64_to_cpu(anchor_header->Primary_Header_LBA);
	if (md_read_blocks(blkdev, ddf_primary_base, 1,
			   (uint8_t *)di->hdr_pri) != 0) {
		printk("ddf_get_disk_info: cannot read primary block\n");
		ddf_free_disk_info(di);
		return (-EIO);
	}

	/* Copy the primary header info the disk_info struct so that we can
	 * reuse the buffer.
	 */
	if ((le32_to_cpu(di->hdr_pri->Signature) != DDF_HEADER_SIGNATURE) ||
	    (le32_to_cpu(di->hdr_pri->Header_Type) != DDF_HEADER_PRIMARY)) {
		printk("ddf primary header signature does not match\n");
		ddf_free_disk_info(di);
		return (-EINVAL);
	}

	/* Read the Physical Disk Data record for this drive so we know what
	 * to reference in the rest of the metadata.
	 */
	offset = le32_to_cpu(di->hdr_pri->pdd_section);
	length = le32_to_cpu(di->hdr_pri->pdd_length);

	/* We could check every section pointer, but it's a good bet that
	 * if one is initialized, then all are initialized.
	 */
	if ((offset == 0xffffffff) || (length == 0xffffffff)) {
		printk("DDF: PDD fields unitialized\n");
		ddf_free_disk_info(di);
		return (-EINVAL);
	}

	di->pdd = kmalloc(length * 512, GFP_KERNEL);
	if (di->pdd == NULL) {
		printk("md_metadata_ddf: failed to allocate buffer\n");
		ddf_free_disk_info(di);
		return (-ENOMEM);
	}
		
	if (md_read_blocks(blkdev, ddf_primary_base + offset, length,
			   (uint8_t *)di->pdd) != 0) {
		printk("ddf_get_disk_info: cannot read pdd\n");
		ddf_free_disk_info(di);
		return (-EIO);
	}

	if (le32_to_cpu(di->pdd->Signature) != DDF_PDD_SIGNATURE) {
		printk("DDF: pdd signature does not match\n");
		ddf_free_disk_info(di);
		return (-EINVAL);
	}

	/*
	 * Ok, we know which disk we are.  Now figure out the rest of the
	 * topology
	 */
	offset = di->hdr_pri->pdr_section;
	length = di->hdr_pri->pdr_length;
	di->pdr = kmalloc(length * 512, GFP_KERNEL);
	if (di->pdr == NULL) {
		printk("md_metadata_ddf: failed to allocate buffer\n");
		ddf_free_disk_info(di);
		return (-ENOMEM);
	}
		
	if (md_read_blocks(blkdev, ddf_primary_base + offset, length,
			   (uint8_t *)di->pdr) != 0) {
		printk("ddf_get_disk_info: cannot read pdr\n");
		ddf_free_disk_info(di);
		return (-EIO);
	}

	if (le32_to_cpu(di->pdr->Signature) != DDF_PDR_SIGNATURE) {
		printk("ddf pdr signature does not match\n");
		ddf_free_disk_info(di);
		return (-EINVAL);
	}

	/* Find the correct PD Entry for our drive */
	di->pde = NULL;
	pdref = le32_to_cpu(di->pdd->PD_Reference);
	for (i = 0; i < le16_to_cpu(di->pdr->Populated_PDEs); i++) {
		di->pde = &di->pdr->entry[i];
		if (le32_to_cpu(di->pde->PD_Reference) == pdref)
			break;
		di->pde = NULL;
	}

	if (di->pde == NULL) {
		printk("ddf_get_disk_info: Cannot find pde for disk!\n");
		ddf_free_disk_info(di);
		return (-EINVAL);
	}

	printk("ddf_get_disk_info: found disk ref= %d, type= 0x%x, state= "
	       "0x%x, size= %lld\n", le32_to_cpu(di->pde->PD_Reference),
	       le16_to_cpu(di->pde->PD_Type), le16_to_cpu(di->pde->PD_State),
	       le64_to_cpu(di->pde->Configured_Size));

	/* Check to see if we already know about this disk.  If the GUIDs are
	 * the same and the kdev_t's are the same, just free the existing
	 * disk_info and return the cached object.
	 */
	list_for_each_entry(tmp_info, &diskqueue, ddf_disk_links) {
		if (memcmp(&di->pde->PD_GUID[0],
			   &tmp_info->pde->PD_GUID[0], 20) == 0) {
			if (tmp_info->blkdev != di->blkdev) {
				printk("Duplicated DDF disk, bailing out\n");
				return (-EINVAL);
				ddf_free_disk_info(di);
			}
			ddf_free_disk_info(di);
			*disk_info = tmp_info;
			return (0);
		}
	}

	list_add(&di->ddf_disk_links, &diskqueue);

	/*
	 * Get the rest of the interesting fields from DDF block.
	 */
	offset = le32_to_cpu(di->hdr_pri->pdd_section);
	length = le32_to_cpu(di->hdr_pri->pdd_length);

	/* Figure out the configuration record mess.  This painfull for
	 * two reasons:
	 * 1) The spec only says how many records there are, but not
	 *    what kind of records they are.  So you have to inspect
	 *    the signatures of each one.
	 *
	 * 2) Adaptec BIOS does follow the spec with regards to the
	 *    Configuration_Record_Length field, making it even harder
	 *    to figure out what is going on.
	 */

	/* XXX The VDCR might be really big.  This needs to be handled */
	length = le32_to_cpu(di->hdr_pri->cr_length);
	offset = le32_to_cpu(di->hdr_pri->cr_section);
	buffer = kmalloc(length * 512, GFP_KERNEL);
	if (buffer == NULL) {
		ddf_free_disk_info(di);
		return (-ENOMEM);
	}

	di->cr = buffer;
	if (md_read_blocks(blkdev, ddf_primary_base + offset, length,
			   buffer) != 0) {
		printk("md_add_to_array_ddf: cannot read vdcr\n");
		ddf_free_disk_info(di);
		return (-EIO);
	}

	crl = le16_to_cpu(di->hdr_pri->Configuration_Record_Length);
	for (i = 0; i < le16_to_cpu(di->hdr_pri->Max_Partitions); i++) {
		switch (le32_to_cpu(((uint32_t *)(buffer))[0])) {
		case DDF_VDCR_SIGNATURE:

			/* Only assign this pointer once.  We will use
			 * pointer arithmitec to index to the right spot
			 * later on.
			 */
			if (di->vdcr == NULL)
				di->vdcr = (struct ddf_vdc_record *)di->cr;
			
			/* Store the variable-sized arrays from the VDCR */
			switch (crl) {
			case 3:
				di->pds_len = 512;
				di->ss_len = 512;
				break;
			case 7:
				di->pds_len = 1024;
				di->ss_len = 2048;
				break;
			case 25:
				di->pds_len = 4096;
				di->ss_len = 8192;
				break;
			case 97:
				di->pds_len = 16384;
				di->ss_len = 32768;
				break;
			case 2:
			case 0xffff:
				printk("Configuration_Record_Length is invalid,"
				       " assuming ADPT method\n");
				di->pds_len =
				    le16_to_cpu(di->hdr_pri->Max_PD_Entries)*4;
				di->ss_len =
				    le16_to_cpu(di->hdr_pri->Max_PD_Entries)*8;
				break;
			default:
				printk(
				   "Unsupported Configuration_Record_Length\n");
				ddf_free_disk_info(di);
				return (-EINVAL);
			}

			buffer += di->pds_len + di->ss_len +DDF_VDCR_PDS_OFFSET;
			break;
		case DDF_SA_SIGNATURE:
			di->sa = (struct ddf_sa_record *)di->cr;
			if (crl == 0xffff) {
				buffer += (DDF_ADPT_SA_LENGTH * 512);
			} else {
				buffer += (crl * 512);
			}
			break;
		case DDF_VUCR_SIGNATURE:
			di->vuc = (struct ddf_vuc_record *)di->cr;
			if (crl == 0xffff) {
				buffer += (DDF_ADPT_VUCR_LENGTH * 512);
			} else {
				buffer += (crl * 512);
			}
			break;
		default:
			printk("DDF: Unknown configuration record\n");
			ddf_free_disk_info(di);
			return (-EINVAL);
		}
	}

	/* If there is a VDCR record, then get the VDR also */
	if (di->vdcr != NULL) {

		di->vdr = kmalloc(DDF_ADPT_VDR_SIZE * 512, GFP_KERNEL);
		if (di->vdr == NULL) {
			ddf_free_disk_info(di);
			return (-ENOMEM);
		}
		offset = le32_to_cpu(di->hdr_pri->vdr_section);
		length = le32_to_cpu(di->hdr_pri->vdr_length);
		if (md_read_blocks(blkdev, ddf_primary_base + offset, length,
				   (uint8_t *)di->vdr) != 0) {
			printk("md_add_to_array_adaptec: cannot read vdr\n");
			ddf_free_disk_info(di);
			return (-EIO);
		}
		if (le32_to_cpu(di->vdr->Signature) !=DDF_VD_RECORD_SIGNATURE) {
			printk("ddf_vdr signature does not match\n");
			ddf_free_disk_info(di);
			return (-EINVAL);
		}
	}

	*disk_info = di;
	return (0);
}

static int
ddf_get_array_info(struct md_ddf_disk_info *di,
		   struct md_ddf_array_info **array_info)
{
	struct md_ddf_array_info *ai;
	int i;

	/* Do we already know about this array? */
	list_for_each_entry(ai, &arrayqueue, ddf_array_links) {
		if (memcmp(&ai->vdcr->VD_GUID[0], &di->vdcr->VD_GUID[0],
			   20) == 0) {
			*array_info = ai;
			return (0);
		}
	}	

	/* This is a new array.  Discover its details */
	ai = kmalloc(sizeof(struct md_ddf_array_info), GFP_KERNEL);
	if (ai == NULL) {
		printk("md_add_to_array_ddf: cannot allocate array_info\n");
		return (-EINVAL);
	}
	memset(ai, 0, sizeof(struct md_ddf_array_info));
	ai->vdcr = di->vdcr;
	ai->vdr = di->vdr;
	ai->DDF_Sequence_Number = le32_to_cpu(di->hdr_pri->Sequence_Number);

	/* Find the VD Entry in the record */
	ai->vde = NULL;
	for (i = 0; i < le16_to_cpu(ai->vdr->Populated_VDEs); i++) {
		ai->vde = &di->vdr->entry[i];
		if (memcmp(&ai->vde->VD_GUID[0], &ai->vdcr->VD_GUID[0], 20) ==0)
			break;
		ai->vde = NULL;
	}

	if (ai->vde == NULL) {
		printk("md_add_to_array_ddf: cannot find VDE\n");
		kfree(ai);
		*array_info = NULL;
		return (-EINVAL);
	}

	ai->pds_array =
	    (uint32_t *)&((uint8_t *)(ai->vdcr))[DDF_VDCR_PDS_OFFSET];
	ai->ss_array =
	    (uint32_t *)&((uint8_t *)(ai->vdcr))[DDF_VDCR_PDS_OFFSET +
						 di->pds_len];

	printk("FOUND NEW DDF ARRAY: PRL= %d RLQ= %d, Block_Count= "
	       "%llu, Size= %llu, Stripe Size= %d\n",
	       ai->vdcr->Primary_RAID_Level, ai->vdcr->RLQ,
	       ai->vdcr->Block_Count, ai->vdcr->Size,
	       (1 << ai->vdcr->Stripe_Size) * 512);

	list_add(&ai->ddf_array_links, &arrayqueue);

	*array_info = ai;

	return (0);
}

static int
ddf_add_spare(struct md_ddf_disk_info *di)
{
	return (0);
}

static void
ddf_free_disk_info(struct md_ddf_disk_info *di)
{
	if (di == NULL)
		return;

	if (di->hdr_anchor != NULL)
		kfree(di->hdr_anchor);
	if (di->hdr_pri != NULL)
		kfree(di->hdr_pri);
	if (di->cd != NULL)
		kfree(di->cd);
	if (di->pdr != NULL)
		kfree(di->pdr);
	if (di->vdr != NULL)
		kfree(di->vdr);
	if (di->cr != NULL)
		kfree(di->cr);
	if (di->pdd != NULL)
		kfree(di->pdd);
	if (di->bbm != NULL)
		kfree(di->bbm);
	if (di->vendor != NULL)
		kfree(di->vendor);
	if (di->blkdev != NULL)
		md_release_blkdev(di->blkdev);
	kfree(di);
}

static int ddf_check_runnable(mddev_t *mddev)
{
	struct md_ddf_array_info *ddf_array;
	uint8_t vde_state;
	int state;

	ddf_array = to_member(mddev)->metadata_private;
	vde_state = ddf_array->vde->VD_State;

	switch (vde_state & DDF_VDE_STATE_MASK) {
	case DDF_VDE_OPTIMAL:
		state = MDK_MSTATE_OPTIMAL;
		break;
	case DDF_VDE_DEGRADED:
		state = MDK_MSTATE_DEGRADED;
		break;
	case DDF_VDE_DELETED:
	case DDF_VDE_MISSING:
	case DDF_VDE_FAILED:
	default:
		return (MDK_MSTATE_FAILED);
	}

	/*
	 * The previous switch statement gave us a hint as to what the metadata
	 * thinks the state is.  Now confirm it by looking at how many disks
	 * we actually have.
	 */
	if (mddev->raid_disks < mddev->max_disks)
		state = MDK_MSTATE_DEGRADED;
	if (mddev->raid_disks < ddf_array->needed_disks)
		state = MDK_MSTATE_FAILED;
	return (state);
}

static int ddf_to_md_disk_state(uint8_t pd_state)
{
	switch (pd_state) {
	case DDF_PDE_ONLINE:
		return (MDK_MSTATE_OPTIMAL);
	case DDF_PDE_REBUILD:
	case DDF_PDE_TRANSITION:
		return (MDK_MSTATE_REBUILD_TARGET);
	case DDF_PDE_FAILED:
	case DDF_PDE_PFA:
	case DDF_PDE_UNRECOVERED:
	case DDF_PDE_MISSING:
	default:
		return (MDK_MSTATE_FAILED);
	}
}

static int ddf_open_disks(mddev_t *mddev, int open)
{
	mdk_member_t *disk;
	struct list_head *tmp;
	struct ddf_header *hdr;
	struct md_ddf_array_info *array;
	struct md_ddf_disk_info *di;
	uint64_t ddf_primary_base;
	uint32_t hdr_seq;
	uint8_t *buffer;
	u_long flags;
	int retval;

	array = mddev->member.metadata_private;
	retval = 0;

	buffer = kmalloc(DDF_BUFSIZ, GFP_KERNEL);
	if (buffer == NULL)
		return (-ENOMEM);
	memset(buffer, 255, DDF_BUFSIZ);

	/* Increment the sequence number, don't let it become -1 or 0 */
	hdr_seq = array->DDF_Sequence_Number;
	if (hdr_seq != 0xffffffff) {
		hdr_seq++;
		if (hdr_seq == 0xffffffff)
			hdr_seq = 1;
	}
	array->DDF_Sequence_Number = hdr_seq;

	ITERATE_CHILDREN(to_member(mddev), disk, tmp, flags) {
		di = disk->metadata_private;
		hdr = di->hdr_pri;
		hdr->Open_Flag = open;
		hdr->Sequence_Number = cpu_to_le32(hdr_seq);
		memcpy(buffer, hdr, sizeof(struct ddf_header));
		ddf_primary_base = le64_to_cpu(hdr->Primary_Header_LBA);
		if (md_write_blocks(to_rdev(disk)->bdev, ddf_primary_base, 1,
				    buffer) < 0) {
			printk("DDF: Error opening DDF\n");
			retval = -1;
			member_put(disk);
			break;
		}
	}

	kfree(buffer);
	return (retval);
}

static int ddf_update_disk(mddev_t *mddev, mdk_rdev_t *disk)
{
	struct list_head *tmp;
	mdk_member_t *target_disk;
	struct md_ddf_disk_info *di, *target_di;
	struct ddf_pd_record *target_pdr;
	struct ddf_pd_entry *pde, *target_pde;
	struct ddf_header *target_hdr;
	uint8_t *buffer;
	uint64_t ddf_primary_base;
	u_long flags;
	int i;

	di = disk->member.metadata_private;
	pde = di->pde;

	if (ddf_open_disks(mddev, 1 /* open */) < 0)
		return (-1);

	buffer = kmalloc(DDF_BUFSIZ, GFP_KERNEL);
	if (buffer == NULL)
		goto bailout;
	memset(buffer, 255, DDF_BUFSIZ);

	/* When updating the PDE, all disks in the set must be updated */
	ITERATE_CHILDREN(to_member(mddev), target_disk, tmp, flags) {
		printk("Updating disk %d\n", target_disk->priority);
		target_di = target_disk->metadata_private;
		target_hdr = target_di->hdr_pri;
		ddf_primary_base = le64_to_cpu(target_hdr->Primary_Header_LBA);

		/* Read in the PDR so we can update the PDE inside of it */
		if (md_read_blocks(to_rdev(target_disk)->bdev,
				   ddf_primary_base +
				   le32_to_cpu(target_hdr->pdr_section),
				   le32_to_cpu(target_hdr->pdr_length), buffer)
				   != 0) {
			printk("md_check_for_sb_ddf: cannot read pdr\n");
			member_put(target_disk);
			goto bailout;
		}

		target_pdr = (struct ddf_pd_record *)buffer;
		if (le32_to_cpu(target_pdr->Signature) != DDF_PDR_SIGNATURE) {
			printk("ddf pdr signature does not match\n");
			member_put(target_disk);
			goto bailout;
		}

		/* Find the PDE */
		target_pde = NULL;
		for (i = 0; i < le16_to_cpu(target_pdr->Populated_PDEs); i++) {
			target_pde = &target_pdr->entry[i];
			if (le32_to_cpu(target_pde->PD_Reference) ==
			    le32_to_cpu(pde->PD_Reference))
				break;
			target_pde = NULL;
		}

		if (target_pde == NULL) {
			member_put(target_disk);
			goto bailout;
		}

		/* Update the PDE */
		memcpy(target_pde, pde, sizeof(struct ddf_pd_entry));

		/* Write back the PDR */
		md_write_blocks(to_rdev(target_disk)->bdev,
				ddf_primary_base +
				le32_to_cpu(target_hdr->pdr_section),
				le32_to_cpu(target_hdr->pdr_length), buffer);
	}

	ddf_open_disks(mddev, 0 /* close */);
bailout:
	if (buffer != NULL)
		kfree(buffer);

	return (0);
}

static int
ddf_update_array(mddev_t *mddev)
{
	return (0);
}

static int
md_set_disk_state_ddf(mddev_t *mddev, mdk_rdev_t *disk, mdk_mstate_t state)
{
	struct md_ddf_disk_info *di;
	int ddf_state;

	ddf_state = md_to_ddf_disk_state(state);
	di = to_member(disk)->metadata_private;

	if (di->pde->PD_State == ddf_state)
		return (0);

	di->pde->PD_State = ddf_state;
	ddf_update_disk(mddev, disk);

	if (ddf_state == DDF_PDE_FAILED) {
		mddev->raid_disks--;
		state = ddf_check_runnable(mddev);
		if (state != to_member(mddev)->state)
			md_set_array_state_ddf(mddev, state);
	}

	return (0);
}

static int
md_set_array_state_ddf(mddev_t *mddev, mdk_mstate_t state)
{
	struct md_ddf_array_info *ddf_array;
	int ddf_state;

	ddf_array = to_member(mddev)->metadata_private;
	ddf_state = md_to_ddf_array_state(state);

	if (ddf_array->vde->VD_State == ddf_state)
		return (0);

	ddf_array->vde->VD_State = ddf_state;
	ddf_update_array(mddev);

	return (0);
}

#if 0
static int
md_set_array_operation_ddf(struct mddev *mddev, struct md_raid_array *array,
			   md_array_operation op)
{
	return (0);
}

static int
md_set_checkpoint_ddf(struct mddev *mddev, struct md_raid_array *array,
		      uint64_t checkpoint)
{
	return (0);
}

static int
md_mark_clean_ddf(struct mddev *mddev)
{
	return (0);
}

static int
md_mark_dirty_ddf(struct mddev *mddev)
{
	return (0);
}

static int
md_array_dirty_ddf(struct mddev *mddev)
{
	return (0);
}

static int
md_delete_array_ddf(struct mddev *mddev)
{
	return (0);
}

static int
md_get_array_operation(struct mddev *mddev, struct md_raid_array *array,
		       md_array_operation *op)
{
	struct md_raid_disk *disk;
	struct md_ddf_array_info *ddf_array;

	ddf_array = array->metadata_private;

	*op = MD_ARRAY_OP_NONE;

	if (ddf_array->vde.VD_State & DDF_VDE_DIRTY) {
		*op = MD_ARRAY_OP_RESYNC;
		return (0);
	}
	
	TAILQ_FOREACH(disk, &mddev->raidarray->disklist, raiddisk_links) {
		if (disk->state == MD_DISK_STATE_TARGET) {
			*op = MD_ARRAY_OP_RESYNC;
			break;
		}
	}

	return (0);
}

static int
md_metadata_sync_ddf(struct mddev *mddev)
{
	return (0);
}

static md_array_state
ddf_to_md_array_state(uint8_t vd_state)
{
	switch (vd_state & DDF_VDE_STATE_MASK) {
	case DDF_VDE_OPTIMAL:
		return (MD_ARRAY_OPTIMAL);
	case DDF_VDE_DEGRADED:
		return (MD_ARRAY_DEGRADED);
	case DDF_VDE_DELETED:
	case DDF_VDE_FAILED:
	case DDF_VDE_MISSING:
	default:
		return (MD_ARRAY_DEAD);
	}
}

#endif
static uint8_t
md_to_ddf_disk_state(mdk_mstate_t state)
{
	switch (state) {
	case MDK_MSTATE_OPTIMAL:
		return (DDF_PDE_ONLINE);
	case MDK_MSTATE_REBUILD_TARGET:
		return (DDF_PDE_REBUILD);
	case MDK_MSTATE_FAILED:
	default:
		return (DDF_PDE_FAILED);
	}
}

static uint8_t
md_to_ddf_array_state(mdk_mstate_t state)
{
	switch (state) {
	case MDK_MSTATE_OPTIMAL:
		return (DDF_VDE_OPTIMAL);
	case MDK_MSTATE_DEGRADED:
		return (DDF_VDE_DEGRADED);
	case MDK_MSTATE_FAILED:
	default:
		return (DDF_VDE_FAILED);
	}
}

static uint32_t ddf_calc_cksum(void *data, int size)
{
	return (0xffffffff);
}

static void ddf_create_guid(char *data, int type)
{
	return;
}

static struct ddf_header * init_header(int header_type, sector_t size)
{
	struct ddf_header *header;
	uint64_t ddf_start;

	ddf_start = size - 1 - 11;
	header = kmalloc(DDF_HEADER_LENGTH, GFP_KERNEL);
	if (header == NULL)
		return (NULL);

	memset(header, 0xff, DDF_HEADER_LENGTH);
	header->Signature = cpu_to_le32(DDF_HEADER_SIGNATURE);
	ddf_create_guid(header->DDF_Header_GUID, 0);
	memcpy(header->DDF_rev, DDF_REV, 8);
	header->Sequence_Number = cpu_to_le32(0x1);
	header->TimeStamp = 0x0;	/* XXX */
	header->Open_Flag = 0x01;
	header->Foreign_Flag = 0x0;
	header->Diskgrouping = 0x0;
	header->Primary_Header_LBA = cpu_to_le64(ddf_start);
	header->Header_Type = header_type;
	header->WorkSpace_Length = cpu_to_le32(0);
	header->Max_PD_Entries = MD_DDF_MAX_PDE;
	header->Max_VD_Entries = MD_DDF_MAX_VDE;
	header->Max_Partitions = cpu_to_le16(0x1);
	header->Configuration_Record_Length = cpu_to_le16(0xffff); /*XXX*/
	header->cd_section = cpu_to_le32(DDF_ADPT_CD_OFFSET);
	header->cd_length = cpu_to_le32(DDF_ADPT_CD_SIZE);
	header->pdr_section = cpu_to_le32(DDF_ADPT_PDR_OFFSET);
	header->pdr_length = cpu_to_le32(DDF_ADPT_PDR_SIZE);
	header->vdr_section = cpu_to_le32(DDF_ADPT_VDR_OFFSET);
	header->vdr_length = cpu_to_le32(DDF_ADPT_VDR_SIZE);
	header->cr_section = cpu_to_le32(DDF_ADPT_VDCR_OFFSET);
	header->cr_length = cpu_to_le32(DDF_ADPT_VDCR_SIZE);
	header->pdd_section = cpu_to_le32(DDF_ADPT_PDD_OFFSET);
	header->pdd_length = cpu_to_le32(DDF_ADPT_PDD_SIZE);
	header->Vendor_Specific_Logs = cpu_to_le32(DDF_ADPT_VENDOR_OFFSET);
	header->Vendor_Specific_Logs_Length = cpu_to_le32(DDF_ADPT_VENDOR_SIZE);

	header->CRC = ddf_calc_cksum(header, DDF_HEADER_LENGTH);

	return (header);
}

static struct ddf_cd_record * init_cd(void)
{
	struct ddf_cd_record *cd;

	cd = kmalloc(DDF_ADPT_CD_SIZE * 512, GFP_KERNEL);
	if (cd == NULL)
		return (NULL);

	memset(cd, 0xff, DDF_ADPT_CD_SIZE * 512);
	cd->Signature = cpu_to_le32(DDF_CONTROLLER_DATA_SIGNATURE);
	ddf_create_guid(cd->Controller_GUID, 0);

	cd->CRC = ddf_calc_cksum(cd, DDF_ADPT_CD_SIZE * 512);

	return (cd);
}

static struct ddf_pd_record * init_pdr(int populated_pdes)
{
	struct ddf_pd_record *pdr;

	pdr = kmalloc(DDF_ADPT_PDR_SIZE * 512, GFP_KERNEL);
	if (pdr == NULL)
		return (NULL);

	memset(pdr, 0xff, DDF_ADPT_PDR_SIZE * 512);
	pdr->Signature = cpu_to_le32(DDF_PDR_SIGNATURE);
	pdr->Populated_PDEs = cpu_to_le16(populated_pdes);
	pdr->Max_PD_Entries = cpu_to_le16(MD_DDF_MAX_PDE);

	pdr->CRC = ddf_calc_cksum(pdr, DDF_ADPT_PDR_SIZE * 512);

	return (pdr);
}

static struct ddf_pd_entry * init_pde(struct ddf_pd_record *pdr, int entry,
				      int type, int state, uint64_t size)
{
	struct ddf_pd_entry *pde;

	pde = &pdr->entry[entry];
	ddf_create_guid(pde->PD_GUID, 0);
	pde->PD_Reference = 0x01;	/* XXX */
	pde->PD_Type = cpu_to_le16(type);
	pde->PD_State = cpu_to_le16(state);
	pde->Configured_Size = cpu_to_le64(size);

	pdr->CRC = ddf_calc_cksum(pdr, DDF_ADPT_PDR_SIZE * 512);

	return (pde);
}

static struct ddf_vd_record * init_vdr(int populated_vdes)
{
	struct ddf_vd_record *vdr;

	vdr = kmalloc(DDF_ADPT_VDR_SIZE * 512, GFP_KERNEL);
	if (vdr == NULL)
		return (NULL);

	memset(vdr, 0xff, DDF_ADPT_VDR_SIZE * 512);
	vdr->Signature = cpu_to_le32(DDF_VD_RECORD_SIGNATURE);
	vdr->Populated_VDEs = cpu_to_le16(populated_vdes);
	vdr->Max_VD_Entries = cpu_to_le16(MD_DDF_MAX_VDE);

	vdr->CRC = ddf_calc_cksum(vdr, DDF_ADPT_VDR_SIZE * 512);

	return (vdr);
}

static struct ddf_vd_entry * init_vde(struct ddf_vd_record *vdr, int entry,
				      int type, int state, int init_state,
				      char *name)
{
	struct ddf_vd_entry *vde;

	vde = &vdr->entry[entry];
	ddf_create_guid(vde->VD_GUID, 0);
	vde->VD_Number = cpu_to_le16(entry);
	vde->VD_Type = cpu_to_le16(type);
	vde->VD_State = cpu_to_le16(state);
	vde->Init_State = cpu_to_le16(init_state);
	memcpy(vde->VD_Name, name, 16);

	vdr->CRC = ddf_calc_cksum(vdr, DDF_ADPT_VDR_SIZE * 512);
	return (vde);
}

static struct ddf_vdc_record * init_vdcr(int count, int ss, int level,
					       int rlq, uint64_t size)
{
	struct ddf_vdc_record *vdcr;

	vdcr = kmalloc(DDF_ADPT_VDCR_SIZE * 512i, GFP_KERNEL);
	if (vdcr == NULL)
		return (NULL);

	memset(vdcr, 0xff, DDF_ADPT_VDCR_SIZE * 512);
	vdcr->Signature = cpu_to_le32(DDF_VDCR_SIGNATURE);
	ddf_create_guid(vdcr->VD_GUID, 0);
	vdcr->Timestamp = 0x0;
	vdcr->Sequence_Number = cpu_to_le32(0x0);
	vdcr->Primary_Element_Count = cpu_to_le16(count);
	vdcr->Stripe_Size = ffs(ss >> 9);
	vdcr->Primary_RAID_Level = level;
	vdcr->RLQ = rlq;
	vdcr->Secondary_Element_Count = 1;
	vdcr->Block_Count = cpu_to_le64(size);
	vdcr->Size = cpu_to_le64(size);
	vdcr->Cache_Flags = cpu_to_le64(0x0);
	vdcr->BG_Rate = 0x0;

	vdcr->CRC = ddf_calc_cksum(vdcr, DDF_ADPT_VDCR_SIZE * 512);

	return (vdcr);
}

static struct ddf_pdd_record * init_pdd(char *guid, uint32_t ref)
{
	struct ddf_pdd_record *pdd;

	pdd = kmalloc(DDF_ADPT_PDD_SIZE * 512, GFP_KERNEL);
	if (pdd == NULL)
		return (NULL);

	memset(pdd, 0xff, DDF_ADPT_PDD_SIZE * 512);
	pdd->Signature = cpu_to_le32(DDF_PDD_SIGNATURE);
	memcpy(pdd->PD_GUID, guid, 24);
	pdd->PD_Reference = cpu_to_le32(ref);
	pdd->Forced_Ref_Flag = 0;
	pdd->Forced_PD_GUID_Flag = 0;

	pdd->CRC = ddf_calc_cksum(pdd, DDF_ADPT_PDD_SIZE * 512);

	return (pdd);
}

static int __init ddf_init(void)
{
	return register_md_metapersonality(&md_ddf_pers);
}

static void ddf_exit(void)
{
	unregister_md_metapersonality(&md_ddf_pers);
}

module_init(ddf_init);
module_exit(ddf_exit);
MODULE_LICENSE("Dual BSD/GPL");
